# Relative import
import sys, pathlib
sys.path.append(str(pathlib.Path(__file__).parent.parent.parent.resolve()))
from sail import cli

import io
import click
import unittest
import subprocess
import requests
from click.testing import CliRunner
from unittest.mock import Mock, patch

# Some commands use os.execlp to pass control to the
# child process, which stops test execution after the
# spawned process exits. We can mock it and just pipe
# to a regular subprocess.
def _execlp_side_effect(*args, **kwargs):
	print(subprocess.check_output(args[1:], encoding='utf8'))

_execlp = Mock(side_effect=_execlp_side_effect)

@patch('os.execlp', _execlp)
@unittest.skipIf(__name__ != '__main__', 'Slow test, must run explicitly')
class TestEnd2End(unittest.TestCase):
	def setUp(self):
		pass

	def test_end2end(self):
		runner = CliRunner(mix_stderr=False)

		with runner.isolated_filesystem():
			result = runner.invoke(cli, ['config', 'api-base', 'http://127.0.0.1:5000/api/1.0/'])
			self.assertEqual(result.exit_code, 0)
			self.assertIn('Option api-base set', result.output)

			# Init
			result = runner.invoke(cli, ['init'])
			print(result.output)
			self.assertIn('Success. The ship has sailed!', result.output)
			self.assertEqual(result.exit_code, 0)

			result = runner.invoke(cli, ['wp', 'option', 'get', 'home'])
			print(result.output)
			self.assertEqual(result.exit_code, 0)
			home = result.output.strip()

			# Import some media
			result = runner.invoke(cli, ['wp', 'media', 'import', 'https://s.w.org/style/images/wp-header-logo.png'])
			print(result.output)
			self.assertEqual(result.exit_code, 0)
			self.assertIn('Imported file', result.output)

			# Download
			result = runner.invoke(cli, ['download', '-y'])
			print(result.output)
			self.assertEqual(result.exit_code, 0)

			# Uploaded file shouldn't be here as we skip uploads by default
			found = subprocess.check_output(['find', '.', '-type', 'f', '-name', 'wp-header-logo.png'], encoding='utf8')
			self.assertNotIn('wp-header-logo.png', found)

			# With uploads now
			result = runner.invoke(cli, ['download', '-y', '--with-uploads'])
			print(result.output)
			self.assertEqual(result.exit_code, 0)

			# File should now be downloaded.
			found = subprocess.check_output(['find', '.', '-type', 'f', '-name', 'wp-header-logo.png'], encoding='utf8')
			self.assertIn('wp-header-logo.png', found)

			# Test PHP
			with open('test.php', 'w') as f:
				f.write('<?php\necho "%s";' % home)

			# Deploy the test file
			result = runner.invoke(cli, ['deploy'])
			print(result.output)
			self.assertEqual(result.exit_code, 0)

			response = requests.get('%s/test.php' % home)
			self.assertTrue(response.ok)
			self.assertEqual(home, response.text)

			# Clean up
			result = runner.invoke(cli, ['destroy', '-y'])
			print(result.output)
			self.assertEqual(result.exit_code, 0)

if __name__ == '__main__':
	unittest.main()



from sail import cli, util

import click, pathlib, json

@cli.command('config')
@click.argument('name', required=True, nargs=1)
@click.argument('value', required=False)
@click.option('--delete', is_flag=True)
def config(name, value=None, delete=False):
	'''Set reusable config variables'''
	valid_names = [
		'provider-token',
		'email',
		'api-base',
	]

	if name not in valid_names:
		raise click.ClickException('Invalid config name: %s' % name)

	filename = (pathlib.Path.home() / '.sail-defaults.json').resolve()
	data = {}

	try:
		with open(filename) as f:
			data = json.load(f)
	except:
		pass

	if not value and not delete:
		if name not in data:
			raise click.ClickException('The option is not set')

		click.echo(data[name])
		return

	if value and delete:
		raise click.ClickException('The --delete flag does not expect an option value')

	if delete and name not in data:
		raise click.ClickException('The option is not set, nothing to delete')

	if delete:
		del data[name]
		click.echo('Option %s deleted' % name)
	else:
		data[name] = value
		click.echo('Option %s set' % name)

	with open(filename, 'w+') as f:
		json.dump(data, f, indent='\t')



from sail import cli, util

import webbrowser
import click

@cli.command()
def admin():
	'''Open your default web browser to the wp-login.php location of your site'''
	root = util.find_root()
	sail_config = util.get_sail_config()
	webbrowser.open(sail_config['login_url'])



from sail import cli, util

import os, subprocess
import click

@cli.command()
@click.option('--root', '--host', is_flag=True, help='Login to the host (not the container) as the root user')
def ssh(root):
	'''Open an interactive SSH shell to the production container or host'''
	as_root = root
	root = util.find_root()
	sail_config = util.get_sail_config()

	# Container as www-data, or host as root
	command = ''
	if not as_root:
		command = 'docker exec -it sail sudo -u www-data bash -c "cd ~/public; bash"'

	click.echo('Spawning an interactive SSH shell for %s.sailed.io' % sail_config['app_id'])

	os.execlp('ssh', 'ssh', '-t',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
		'root@%s.sailed.io' % sail_config['app_id'],
		command
	)



from sail import cli, util

import os, subprocess
import click, shlex

@cli.command(context_settings=dict(ignore_unknown_options=True))
@click.argument('command', nargs=-1)
def wp(command):
	'''Run a WP-CLI command on the production host'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	command = shlex.join(command)

	click.echo('Spawning SSH and running WP-CLI on %s.sailed.io' % sail_config['app_id'], err=True)

	os.execlp('ssh', 'ssh', '-t',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
		'root@%s.sailed.io' % sail_config['app_id'],
		'docker exec -it sail sudo -u www-data bash -c "cd ~/public; wp %s"' % command
	)



from sail import cli, util

import os, subprocess
import click

@cli.command()
@click.option('--nginx', is_flag=True)
@click.option('--php', is_flag=True)
@click.option('--nginx-access', is_flag=True)
@click.option('--nginx-error', '--nginx-errors', is_flag=True)
@click.option('--php-error', '--php-errors', is_flag=True)
@click.option('--follow', '-f', is_flag=True)
@click.option('--lines', '-n', type=int)
def logs(nginx, php, nginx_access, nginx_error, php_error, follow, lines):
	'''Query and follow logs from the production server'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	click.echo('Querying logs on %s.sailed.io' % sail_config['app_id'])

	settings = []
	if nginx_access or nginx:
		settings.append('-t nginx_access')
	if nginx_error or nginx:
		settings.append('-t nginx_error')
	if php_error or php:
		settings.append('-t php')

	if follow:
		settings.append('--follow')
	elif lines:
		settings.append('--lines %d' % lines)
	else:
		settings.append('--lines 30')

	settings = ' '.join(settings)

	os.execlp('ssh', 'ssh', '-t',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
		'root@%s.sailed.io' % sail_config['app_id'],
		'journalctl --no-hostname %s' % settings
	)



from sail import cli, util

import requests, json, os, subprocess, time
import click, hashlib, pathlib, shutil
from datetime import datetime

@cli.command()
def backup():
	'''Backup your production files and database to your local .backups directory'''
	root = util.find_root()
	sail_config = util.get_sail_config()

	click.echo()
	click.echo('# Backing up')

	app_id = sail_config['app_id']
	backups_dir = pathlib.Path(root + '/.backups')
	backups_dir.mkdir(parents=True, exist_ok=True)
	progress_dir = pathlib.Path(backups_dir / ('.%s.progress' % hashlib.sha256(os.urandom(32)).hexdigest()[:8]))
	(progress_dir / 'www').mkdir(parents=True)
	(progress_dir / 'uploads').mkdir()

	database_filename = 'database.%s.sql.gz' % hashlib.sha256(os.urandom(32)).hexdigest()[:8]

	click.echo('- Downloading application files')

	# Download files FROM production
	p = subprocess.Popen([
		'rsync', ('-rtlv' if util.debug() else '-rtl'),
		'-e', 'ssh -i %s/.sail/ssh.key -o UserKnownHostsFile=%s/.sail/known_hosts -o IdentitiesOnly=yes -o IdentityFile=%s/.sail/ssh.key' % (root, root, root),
		'--filter', '- .*', # Exclude all dotfiles
		'--filter', '- wp-content/debug.log',
		'--filter', '- wp-content/uploads',
		'--filter', '- wp-content/cache',
		'--copy-dest', '%s/' % root,
		'root@%s.sailed.io:/var/www/public/' % app_id,
		'%s/www/' % progress_dir,
	])

	while p.poll() is None:
		util.loader()

	if p.returncode != 0:
		shutil.rmtree(progress_dir)
		raise click.ClickException('An error occurred during backup. Please try again.')

	click.echo('- Downloading uploads')

	# Download files FROM production
	p = subprocess.Popen([
		'rsync', ('-rtlv' if util.debug() else '-rtl'),
		'-e', 'ssh -i %s/.sail/ssh.key -o UserKnownHostsFile=%s/.sail/known_hosts -o IdentitiesOnly=yes -o IdentityFile=%s/.sail/ssh.key' % (root, root, root),
		'--filter', '- .*', # Exclude all dotfiles
		'--copy-dest', '%s/wp-content/uploads/' % root,
		'root@%s.sailed.io:/var/www/uploads/' % app_id,
		'%s/uploads/' % progress_dir,
	])

	while p.poll() is None:
		util.loader()

	if p.returncode != 0:
		shutil.rmtree(progress_dir)
		raise click.ClickException('An error occurred during backup. Please try again.')

	click.echo('- Exporting WordPress database')

	p = subprocess.Popen(['ssh',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
		'root@%s.sailed.io' % sail_config['app_id'],
		'docker exec sail bash -c "mysqldump --quick --single-transaction --default-character-set=utf8mb4 -uroot wordpress | gzip -c9 > /var/www/%s"' % database_filename
	])

	while p.poll() is None:
		util.loader()

	if p.returncode != 0:
		shutil.rmtree(progress_dir)
		raise click.ClickException('An error occurred in SSH. Please try again.')

	click.echo('- Export completed, downloading database')

	p = subprocess.Popen([
		'rsync', ('-tv' if util.debug() else '-t'),
		'-e', 'ssh -i %s/.sail/ssh.key -o UserKnownHostsFile=%s/.sail/known_hosts -o IdentitiesOnly=yes -o IdentityFile=%s/.sail/ssh.key' % (root, root, root),
		'root@%s.sailed.io:/var/www/%s' % (sail_config['app_id'], database_filename),
		'%s/database.sql.gz' % progress_dir
	])

	while p.poll() is None:
		util.loader()

	if p.returncode != 0:
		shutil.rmtree(progress_dir)
		raise click.ClickException('An error occurred in rsync. Please try again.')

	click.echo('- Cleaning up production')

	p = subprocess.Popen(['ssh',
		'-i', '%s/.sail/ssh.key' % root,
		'-o', 'UserKnownHostsFile=%s/.sail/known_hosts' % root,
		'-o', 'IdentitiesOnly=yes',
		'-o', 'IdentityFile=%s/.sail/ssh.key' % root,
		'root@%s.sailed.io' % sail_config['app_id'],
		'docker exec sail rm /var/www/%s' % database_filename
	])

	while p.poll() is None:
		util.loader()

	if p.returncode != 0:
		shutil.rmtree(progress_dir)
		raise click.ClickException('An error occurred in SSH. Please try again.')

	timestamp = datetime.now().strftime('%Y-%m-%d-%H%M%S.tar.gz')
	target = pathlib.Path(backups_dir / timestamp)

	click.echo('- Archiving and compressing backup files')

	p = subprocess.Popen([
		'tar', ('-cvzf' if util.debug() else '-czf'), target.resolve(), '-C', progress_dir.resolve(), '.'
	])

	while p.poll() is None:
		util.loader()

	if p.returncode != 0:
		shutil.rmtree(progress_dir)
		raise click.ClickException('An error occurred during backup. Please try again.')

	shutil.rmtree(progress_dir)

	click.echo('- Backup completed at .backups/%s' % timestamp)
